part of player;

class PlayerProvider with ChangeNotifier {
  final VoidCallback onComplete;
  final ValueChanged<VideoPlayerController> onInit;
  final bool autoPlay;
  Duration _hideControlsIn;
  final bool onlyFullscreen;
  FutureOr<VideoPlayerController> _controller;
  bool _hasError = false;

  bool _isControlsShown = false;

  int _hideControlsMatcher = 0;
  bool _isFullscreen;
  VoidCallback _listener;
  bool _isDisposed = false;

  bool get isControlsShown => _isControlsShown;
  bool get mounted => !_isDisposed;
  VideoPlayerValue get value =>
      controller?.value ?? VideoPlayerValue.uninitialized();
  bool get isFullscreen => _isFullscreen;
  VideoPlayerController get controller =>
      _controller is VideoPlayerController ? _controller : null;
  bool get initialized => controller?.value?.initialized ?? false;

  bool get hasError => controller?.value?.hasError ?? _hasError;

  StreamSubscription<PhoneStateCallEvent> _subscription;
  Future<void> _phoneStateListener(PhoneStateCallEvent event) async {
    if (event.stateC == 'true') {
      await controller.pause();
      _showControls(autoHide: false);
    }
  }

  PlayerProvider({
    @required FutureOr<VideoPlayerController> controller,
    this.onComplete,
    this.onInit,
    this.autoPlay,
    Duration hideControlsIn,
    this.onlyFullscreen,
  })  : _controller = controller,
        _hideControlsIn = hideControlsIn {
    _isFullscreen = onlyFullscreen;
    _listener = () {
      if (value.hasError) {
        notifyListeners();
      } else if (value.isCompleted) {
        _showControls();
        if (onComplete != null) onComplete();
      }
    };
    _subscription = phoneStateCallEvent.listen(_phoneStateListener);
  }

  void init() async {
    if (_isDisposed) return;
    _hasError = false;
    controller?.removeListener(_listener);
    controller?.addListener(_listener);
    notifyListeners();
    try {
      if (controller?.value?.hasError ?? false) {
        controller.value = VideoPlayerValue.uninitialized();
      }
      if (_controller is Future) {
        _controller = await _controller;
      }
      await controller.initialize();
      if (onInit != null) onInit(_controller);
      if (onlyFullscreen) enterFullscreen();
      if (autoPlay) {
        await controller.play();
      }
      Wakelock.enable();
    } catch (err) {
      _hasError = true;
      //IGNORE error may be handled by [listener]
    } finally {
      notifyListeners();
    }
  }

  void changeSource(FutureOr<VideoPlayerController> controller) {
    if (controller is VideoPlayerController &&
        this.controller?.dataSource == controller.dataSource) {
      return;
    }
    this.controller?.dispose();
    this._controller = controller;
    init();
  }

  ///`Play`, `Replay` or `Pause` the video
  Future<void> playPause() async {
    if (value.isPlaying) {
      await controller.pause();
    } else if (value.position == value.duration) {
      await controller.seekTo(Duration.zero);
      await controller.play();
    } else if (!value.isPlaying) {
      await controller.play();
    }
    _showControls();
  }

  ///Seeks [seconds]
  ///
  ///Reverse if [seconds]<0
  ///
  ///Forward if [seconds]>0
  Future<void> seek(int seconds) async {
    final isCompleted = value.isCompleted;
    await controller
        .seekTo(controller.value.position + Duration(seconds: seconds));
    if (isCompleted && seconds < 0) {
      await controller.play();
    }
    if (_isControlsShown) _showControls();
  }

  ///Shows controls for [_visibleDuration] and disables
  void _showControls({bool autoHide = true}) {
    if (_isDisposed) return;
    _isControlsShown = true;
    notifyListeners();
    final matcher = _hideControlsMatcher + 1;
    if (value.isPlaying && autoHide) {
      Future.delayed(
        _hideControlsIn,
        () => _hideControls(matcher),
      );
    }
    _hideControlsMatcher = matcher;
  }

  void _hideControls(int matcher) {
    if (_isDisposed || matcher != _hideControlsMatcher) return;
    if (_isControlsShown) {
      _isControlsShown = false;
      notifyListeners();
    }
  }

  void changeControlsVisibility() {
    log('CHANGE CONTROLS VISIBILITY TO ${!_isControlsShown}', name: 'CONTROLS');
    if (_isControlsShown) {
      _hideControls(_hideControlsMatcher);
    } else {
      _showControls();
    }
  }

  void enterFullscreen() {
    if (_isDisposed) return;
    SystemChrome.setEnabledSystemUIOverlays([]);

    if (value.fullScreenOrientation == Orientation.portrait) {
      SystemChrome.setPreferredOrientations([
        DeviceOrientation.portraitDown,
        DeviceOrientation.portraitUp,
      ]);
    } else {
      SystemChrome.setPreferredOrientations([
        DeviceOrientation.landscapeRight,
        DeviceOrientation.landscapeLeft,
      ]);
    }
    _isFullscreen = true;
    notifyListeners();
  }

  void exitFullscreen() {
    if (_isDisposed) return;
    //TODO Handle orientation like a PRO in future
    SystemChrome.setPreferredOrientations([
      DeviceOrientation.portraitDown,
      DeviceOrientation.portraitUp,
    ]);
    SystemChrome.setEnabledSystemUIOverlays(SystemUiOverlay.values);
    _isFullscreen = false;
    notifyListeners();
  }

  void toggleFullscreen() {
    if (_isFullscreen)
      exitFullscreen();
    else
      enterFullscreen();
  }

  @override
  void dispose() {
    _isDisposed = true;
    _subscription.cancel();
    controller.dispose();
    Wakelock.disable();
    super.dispose();
  }
}
